:fl

C-Tips

praxis

exceptions

@d:Exceptions  in ANSI-C nutzen

@t:Mit einer Handvoll Makros

@a:Peter Simons

@v:Exceptions sind eine der Neuerungen objektorientierter Sprachen wie 
Java und C++. Mit wenig Code und einigen <I>#define<I>s knnen auch
C-Programmierer diese moderne Form der Fehlerbehandlung einsetzen.

@$:Der Hauptsatz der modularen Programmierung besagt, da eine groe
Aufgabe sich leichter lsen lt, wenn man sie in mehrere kleine,
mglichst unabhngige Teile zerlegt und diese einzeln bearbeitet.
Dies geschieht in Form von Unterroutinen, die eine bestimmte, berschaubare
Funktion ausfhren.

Hat man passende Unterroutinen geschrieben, besteht die
Erstellung des Hauptprogramms im wesentlichen daraus, diese
zu einem Ganzen zu verbinden, so da man das groe Problem auf einer
relativ hohen Abstraktionsebene lsen kann.

Typischerweise sieht ein C-Programm ungefhr so aus wie in Listing**1.
Das Hauptprogramm ruft eine Routine auf, die eine Routine aufruft,
die wiederum eine Routine aufruft ...

!!!Hier ungefhr Listing 1

Wie reagiert nun <I>get_file_length()<I>, wenn
es die Datei nicht vorfindet, deren Gre gefragt ist? Eine
sauber geschriebene Funktion wird diesen Umstand bemerken und den
Fehler im Rckgabewert melden, beispielsweise durch den
Wert <I>-1<I> statt einer positiven Dateilnge.

Wenn dies geschieht, hat <I>load_file()<I> den Fehler am
Hals. Da die Datei offensichtlich nicht zu laden ist, kann
die Funktion nicht fortfahren. Mit dem Problem
umgehen kann es aber auch nicht, denn es ist ja nur ein
kleines, unschuldiges Unterprogramm, das auf Auftrag von oben
handelt. Also wird es (hoffentlich) den Fehler erkennen und ihn
seinerseits im Returncode weitergeben.

Bei diesem Programmaufbau mu jede Routine stndig auf der Hut sein,
ob vielleicht in einer anderen etwas nicht geklappt hat; jede
Fehlermglichkeit 
verlangt die Aufmerksamkeit des Programmierers. Bei jedem Aufruf
besteht die Gefahr, da irgendwo ein <I>malloc()<I> fehlschlgt oder
eine Datei nicht existiert, und deshalb das ganze Funktionengebude
kippt.

rgerlich ist dies insbesondere, da die meisten Routinen auf den Fehler
gar nicht reagieren knnen. Wenn die aufgerufene Prozedur scheitert, 
mssen auch sie die Arbeit abbrechen, und um diese Situation berhaupt 
erkennen zu knnen, ist stndiges Testen des Rckgabewerts ntig.

Dieses dauernde Prfen ist nicht nur lstig, es macht den
Code auch schwieriger zu schreiben und zu lesen, da sich dadurch oft
komplizierte Verschachtelungen von <I>if<I>s ergeben und
ein unbersichtlicher Programmflu entsteht.

@z:Das Gegenmittel: Exceptions

@$:Schon vor Jahren erkannten Designer anderer Sprachen diesen Mangel und
stellten dem leidgeprften Programmierer sogenannte
Exceptions zur Verfgung. Die bekanntesten Sprachen mit
diesem Mechanismus drften C++ und Java sein.

Dort "wirft" eine Routine eine Exception, wenn sie einen Fehler
erkennt. Wie der Name sagt, ist dies ein Ausnahmezustand, um den sich
irgendjemand kmmern soll -- aber bitte nicht die Routine selber.
Die Exception wandert nun automatisch die gesamte Kette an Aufrufen
zurck, die zur fehlgeschlagenen Aktion fhrten,
bis eine der Routinen sich fr die Behandlung des Fehlers zustndig erklrt.
Will keine die Exception verarbeiten, bricht das Programm mit
einer Fehlermeldung ab.

Durch Verwendung von Exceptions wre der Code im ersten Beispiel
deutlich einfacher geworden. Nur <I>get_file_length()<I> mu einen
Fehler erkennen. Wenn diese Routine das gewnschte Ergebnis nicht
ermitteln kann, wirft sie eine Exception. <I>load_file()<I> und
<I>LoadConfigFiles()<I> brauchen sich um Fehler nicht zu kmmern. Erst
<I>main()<I> wird die Exception schlielich annehmen und dem Benutzer
melden, da das Laden der Konfigurationsdateien gescheitert ist.

@z:Und wie geht das in C?

@$:Wie erwhnt, kennt ANSI-C keine Exceptions. Allerdings bietet
es zwei Funktionen, mit denen man sie simulieren
kann: <I>setjmp()<I> und <I>longjmp()<I>. Die beiden funktionieren wie 
ein <I>goto<I>, mit der ntzlichen Eigenheit, da sie
ber Routinengrenzen hinweg springen knnen.

<I>setjmp()<I> erwartet einen Parameter vom Typ <I>jmp_buf<I>. In ihm
speichert es den aktuellen Stack-Zeiger und den Programm-Zeiger. Als
Rckgabewert liefert es <I>0<I>. <I>longjmp()<I> kann nun genau an
diese Stelle des Programms zurckspringen -- und zwar von jedem
beliebigen, folgenden Punkt der Programmausfhrung. Es erwartet
ebenfalls einen Parameter vom Typ <I>jmp_buf<I>, der vorher von
<I>setjmp()<I> initialisiert worden sein mu, sowie einen
Integer-Wert.

Der <I>longjmp()<I>-Aufruf schreibt nun die in <I>jmp_buf<I>
gespeicherten Stack- und Programm-Zeiger in die 
CPU-Register zurck. Dadurch luft das Programm an
exakt der Stelle weiter, wo der zu <I>jmp_buf<I> gehrende
<I>setjmp()<I>-Call stattfand. Anders als vorher gibt <I>setjmp()<I>
diesmal allerdings nicht <I>0<I> zurck, sondern den Integer-Parameter des
<I>longjmp()<I>-Calls. Zur Verdeutlichung zeigt Listing**2 ein
kurzes  sinnloses Programm, das dieses Prinzip illustriert.

!!!Hier etwa Listing 2

Es passiert folgendes: Das Hauptprogramm ruft zuerst
<I>setjmp()<I>auf. Dieses kehrt mit dem Wert <I>0<I> zurck, worauf
<I>Routine_A()<I> startet und in in <I>Routine_B()<I> springt. Hier
kommt <I>longjmp()<I> zum Zuge, das den Programmablauf in der
<I>switch<I>-Anweisung von <I>main()<I> fortsetzt -- dieses Mal
allerdings mit dem Wert <I>1<I>, so da das Programm nach einem Gru an
Arno endet.

@z:Im Innern der Ausnahme

@$:Mit diesen beiden Funktionen und einigen Makros ist es jetzt ein
leichtes, das oben beschriebene Exception-Modell zu implementieren.
Zunchst ein paar Worte zum dahinterliegenden Konzept.

Offensichtlich ist die Information "Exception" allein fr eine
sinnvolle Fehlerbehandlung nicht ausreichend. Zumindest mchte man
gerne wissen, was fr eine Art von Fehler aufgetreten ist.
Dabei helfen verschiedene Arten von Exceptions, die der
Integer-Parameter in <I>longjmp()<I> verkrpern kann.

:tb

@lf:enum {
@lf:    NO_EXCEPTION = 0,
@lf:    INVALID_ARGUMENT_EXCEPTION,
@lf:    OUT_OF_MEMORY_EXCEPTION,
@lf:    IO_EXCEPTION,
@lf:    [...]
@lf:};

:fl

@$:Eigene Exceptions kann jeder nach Bedarf hinzufgen.
Weiterhin bekommt das Programm ein zustzliches Modul,
das die Verwaltung der <I>setjmp()<I>-Zustnde bernimmt. Ein
Beispiel zeigt in Listing**3.

!!!Hier etwa Listing 3 (exceptions.c)

Das Modul enthlt ein Array von <I>jmp_buf<I>s, das einen
nicht-dynamischen Stack implementiert. Seine Gre bestimmt die
Schachtelungstiefe der Routinen, die dort mit <I>setjmp()<I> Werte
ablegen. Er sollte also nicht zu klein bemessen sein. Im Beispiel ist
Platz fr 256 Exceptions, was fr bliche Anwendungen 
ausreichen sollte. Die im selben Modul definierte Routine
<I>longjmperror()<I> springt ein, wenn das Programm auf eine nicht
abgefangene Exception trifft.
Die Implementierung der Exceptions enthlt <I>exceptions.h<I> in Listing**4.

!!!Hier etwa Listing 4 (exceptions.h)

Motor des ganzen sind die Prprozessor-Makros <I>TRY, HANDLE, THROW,
OTHERWISE<I> und <I>PASSTHROUGH<I>. <I>TRY<I> ruft <I>setjmp()<I> auf
und speichert den aktuellen Zustand des Programms auf dem
Exception-Stack. Weiterhin erffnet es eine Folge von
<I>if<I>-Statements, die nachher eventuelle Exceptions verarbeiten
sollen. <I>TRY<I> leitet Code ein, der fehlschlagen kann.
Programmtext, dem kein <I>TRY<I> vorangeht, wei nichts von Exceptions
und wird beim Auftreten einer Exception bersprungen.

Mit dem <I>HANDLE<I>-Makro nimmt das Programm eine bestimmte Exception
an und verarbeitet sie. Zur Verdeutlichung ein kurzes
Beispiel:

:tb

@lf:TRY {
@lf:   buffer = load_file(filename);
@lf:}
@lf:HANDLE(IO_EXCEPTION) {
@lf:   error("Couldn't load my config files.");
@lf:}

:fl

@$:<I>THROW<I> lst eine Exception aus und speichert deren Typ
in einer internen, von <I>exceptions.c<I> bereitgestellten
Variablen. Wenn auf dem Exception-Stack gltige Eintrge liegen,
fhrt das Makro <I>longjmp()<I> aus, andernfalls ruft es
<I>longjmperror()<I>. Die Verwendung von <I>THROW<I> ist denkbar
einfach:

@lf:THROW(IO_EXCEPTION);

@$:Zu beachten ist, da <I>THROW<I> mit einem Semikolon abgeschlossen
werden mu. Das liegt daran, da es ein Statement im Sinne von C ist,
whrend <I>HANDLE<I> oder <I>TRY<I> eher der <I>if<I>-Syntax
nachempfunden sind. Sobald eine Routine mit <I>TRY<I> und <I>HANDLE<I>
eine Exception annimmt, bekommt sie alle im <I>TRY<I>-Block
auftretenden Ausnahmen und nicht nur die, die sie auch tatschlich
haben mchte. <I>OTHERWISE<I> und <I>PASSTHROUGH<I> reichen in diesem
Fall die unerwnschten Exceptions nach oben weiter. Zur
Illustration erneut ein Code-Fragment:

:tb

@lf:TRY {
@lf:     buffer = load_file(filename);
@lf:}
@lf:HANDLE(IO_EXCEPTION) {
@lf:     error("Failed to read my config file.");
@lf:}
@lf:OTHERWISE {
@lf:     PASSTHROUGH();
@lf:}

:fl
@z:To handle or not to handle

@$:Nachdem die technischen Voraussetzungen fr Exceptions geschaffen
sind, bleibt noch zu klren, wie man sie sinnvoll
verwendet. Anders gefragt: Welche Routine soll Exceptions
handhaben und welche nicht? Die Antwort ist ein klares "das kommt
drauf an".

Generell sollte nur die Funktion eine Exception annehmen, die den
Fehler beheben oder verarbeiten kann. In der Regel sind dies lediglich
Routinen, die in der Programmhierarchie weit oben
stehen, da sie die Kommunikation mit dem Benutzer steuern. Eine
Unterroutine kann meist nicht wissen, wie mit einem Fehler
umzugehen ist, da sie den Gesamtzusammenhang im Programm nicht
kennt und auch nicht kennen soll.

!!!Hier etwa Listing 5

Trotzdem gibt es Flle, in denen auch eine tiefer im Programmflu
angesiedelte Funktion eine Exception annehmen sollte, selbst wenn sie
sie nicht verarbeiten kann. Dies ist dann ntig, wenn sie im
Fehlerfall Ressourcen freigeben mu. Listing**5 illustriert dies: Zwar
kann die Routine <I>load_file()<I> den einen eventuell auftretenden
Fehler in <I>get_filesize_from_handle()<I> nicht beheben, trotzdem
sollte sie die Exception annehmen, denn es gilt den Dateideskriptor
<I>fh<I> zu schlieen, bevor das Programm weiterluft.

Das alles wre fast zu schn, gbe es nicht noch einen Haken.
Glcklicherweise ist er sehr klein. Der ANSI-Standard sagt ber die
Implementierung von <I>setjmp()<I>, da lediglich der Stack- und der
Programm-Zeiger gespeichert werden mssen, nicht jedoch alle CPU-Register.

Bei Verwendung eines optimierenden C-Compilers kann dies zu folgender
Situation fhren: Eine lokale Variable steht nicht im Speicher,
sondern in einem freien Register der CPU, da dies
effizienter ist. Fhrt man jetzt <I>setjmp()<I> aus, so ist nicht
garantiert, da es den Wert dieser Variablen speichert. Bei
einem spteren <I>longjmp()<I> kann sie also 
einen falschen Wert enthalten.

Deklariert man die Variable als <I>volatile<I>, bannt das die Gefahr,
denn der Compiler ist dann gezwungen, sie auf dem Stack anzulegen. Die
meisten C-Compiler, insbesondere GNUs <I>gcc<I>, erkennen diesen Fall
brigens und geben beim bersetzen eine entsprechende Warnung aus.
Beim <I>gcc<I> mu man diese Warnung durch Angabe des Parameters
<I>-W<I> erst einschalten, standardmig verschweigt er 
diese Fehlerquelle.

Allzu dramatisch ist dieses Problem nicht, vor allem, weil die meisten
<I>setjmp<I>-Implementierungen mehr speichern als die vom
ANSI-Standard verlangten Zeiger. Eine gute Maschinenbeschreibung fr
den <I>gcc<I> kann dies dem C-Compiler sogar mitteilen, damit er
besser entscheiden kann, welche Flle berhaupt gefhrlich sind und
welche nicht. (ck)

@au1:Peter Simons

@au2:arbeitet im Forschungszentrum Informationstechnik (GMD) und
beschftigt sich mit dem Forschungsbereich Netzwerke. Auerdem ist er
Programmierer aus Leidenschaft und Autor zahlreicher Programme aus den
Bereichen Internet und Kryptographie.

!!!BUs

@bu:Ein typisches C-Programm, in dem Routinen sich gegenseitig
aufrufen und jede sich um mgliche Fehler kmmern mu (Listing**1).

@bu:Mit <I>setjmp()<I> und <I>longjmp()<I> lassen sich nicht-lokale
<I>goto<I>-Statements implementieren (Listing**2).

@bu:<I>exceptions.c<I> stellt den Exception-Stack und die
<I>longjmperror()<I> zur Verfgung. Diese Funktion behandelt
nicht abgefangene Exceptions (Listing**3).

@bu:Die Header-Datei <I>exceptions.h<I> enthlt die Nummern der
Execptions und alle zu ihrer Behandlung ntigen Makros (Listing**4).

@bu:<I>OTHERWISE<I> und <I>PASSTHROUGH<I> helfen bei der Freigabe
lokaler Ressourcen im Fehlerfall (Listing**5).


:li

!!! Listing 1
int main(int argc, char ** argv) {
    [...]

    /* Load my config files and insert the values into the
       appropriate structure. */

    config = LoadConfigFiles("/etc/myprogramrc");
    if (config == NULL)
        exit(1);

    [...]
}

struct Configuration *
LoadConfigFiles(char * filename) {
    [...]

    /* Load the file into memory so we can process it. */

    buffer = load_file(filename);
    if (buffer == NULL)
        return NULL;

    [...]
}

char * load_file(char * filename) {
    [...]

    file_size = get_file_length(filename);
    if (file_size == -1)
        return NULL;
    buf = malloc(file_size + 1);

    [...]
}



!!!Listing 2

#include <stdio.h>
#include <setjmp.h>

jmp_buf  env;

int
main(int argc, char ** argv) {
    switch (setjmp(env)) {
      case 0:
          Routine_A();
          break;
      case 1:
          printf("Hallo Arno!\n");
          break;
    }
}

void Routine_A(void) {
    Routine_B();
}

void Routine_B(void) {
    longjmp(env, 1);
}

!!!Listing 3

#include <syslog.h>
#define __SOURCE_COMPILE__
#include "exceptions.h"

jmp_buf        exception_env_stack[MAX_NEST_EXCEPTION];
unsigned int   exception_env_stack_pos = 0;
unsigned int   exception_id = 0;

void
longjmperror(void) {
    syslog(LOG_ERR, "FATAL ERROR: Uncaught exception %d\n", exception_id);
    exit(exception_id);
}


!!!Listing 4

-------------------- exceptions.h --------------------
#ifndef __LIB_EXCEPTIONS_H__
#define __LIB_EXCEPTIONS_H__ 1

#include <setjmp.h>

enum {
    NO_EXCEPTION = 0,
    INVALID_ARGUMENT_EXCEPTION,
    OUT_OF_MEMORY_EXCEPTION,
    IO_EXCEPTION
};

#define TRY \
    exception_id = setjmp(exception_env_stack[exception_env_stack_pos++]); \
    if (exception_id == NO_EXCEPTION)

#define HANDLE(e) \
    else if (exception_id == (e))

#define OTHERWISE \
    else

#define PASSTHROUGH() \
    THROW(exception_id)

#define THROW(e) \
    exception_id = e; \
    if (exception_env_stack_pos == 0) \
       longjmperror(); \
    longjmp(exception_env_stack[--exception_env_stack_pos], (e))

#ifndef MAX_NEST_EXCEPTION
#  define MAX_NEST_EXCEPTION 256
#endif

#ifndef __SOURCE_COMPILE__
extern jmp_buf        exception_env_stack[];
extern unsigned int   exception_env_stack_pos;
extern unsigned int   exception_id;
#endif /* ! __SOURCE_COMPILE__ */

#endif /* !__LIB_EXCEPTIONS_H__ */


!!!Listing 5

char * load_file(char * filename) {
    FILE *   fh = NULL;
    int      size;

    fh = fopen(filename, "r");
    if (fh == NULL) {
        THROW(IO_EXCEPTION);
    }
    TRY {
        size = get_filesize_from_handle(fh);
        [...]
    }
    OTHERWISE {
        if (fh != NULL)
            fclose(fh);
        PASSTHROUGH();
    }
}
